"""
LangGraph 자동 패치
LangGraph의 노드 실행과 conditional edges를 자동으로 trace합니다.
"""

from typing import Any, Callable, Optional, Union, Dict, Sequence
from functools import wraps
import sys


def patch_langgraph():
    """LangGraph를 패치하여 자동으로 노드 실행과 routing을 trace합니다."""
    try:
        import langgraph
        from langgraph.graph import StateGraph
    except ImportError:
        # LangGraph가 설치되지 않은 경우 패치하지 않음
        return

    from ...client import _get_active_client

    # 이미 패치되었는지 확인
    if hasattr(StateGraph.add_node, '_nora_patched'):
        return

    # 원본 메서드 저장
    original_add_node = StateGraph.add_node
    original_add_conditional_edges = StateGraph.add_conditional_edges
    
    # 노드 이름 -> 마지막 execution span ID 매핑 (conditional edge 부모 추적용)
    _node_execution_spans: Dict[str, str] = {}

    def traced_add_node(self, name: str, action: Callable, **kwargs):
        """노드 추가 시 자동으로 trace 데코레이터를 적용합니다."""
        import inspect
        client = _get_active_client()
        
        # 이미 traced된 함수인지 확인
        if client and callable(action) and not hasattr(action, '_nora_traced'):
            original_action = action
            
            # async 함수인지 확인
            if inspect.iscoroutinefunction(original_action):
                # async wrapper
                @wraps(original_action)
                async def async_wrapper(*args, **kwargs):
                    from ...client import _current_execution_span
                    
                    # 현재 execution span context 가져오기
                    current_span = _current_execution_span.get()
                    
                    # 함수 실행 (이 시점에 execution span이 설정되어 있음)
                    result = await original_action(*args, **kwargs)
                    
                    # 실행 후 현재 execution span ID 저장 (conditional edge에서 사용)
                    # 다시 가져와야 함 (trace 데코레이터가 설정한 값)
                    current_span = _current_execution_span.get()
                    if current_span and current_span.get("id"):
                        _node_execution_spans[name] = current_span["id"]
                    
                    return result
                
                # trace 데코레이터 적용 (wrapper를 감쌈)
                traced_action = client.trace(async_wrapper, span_kind="langgraph_node", name=name)
                traced_action._nora_traced = True
                return original_add_node(self, name, traced_action, **kwargs)
            else:
                # sync wrapper
                @wraps(original_action)
                def sync_wrapper(*args, **kwargs):
                    from ...client import _current_execution_span
                    
                    # 현재 execution span context 가져오기
                    current_span = _current_execution_span.get()
                    
                    # 함수 실행 (이 시점에 execution span이 설정되어 있음)
                    result = original_action(*args, **kwargs)
                    
                    # 실행 후 현재 execution span ID 저장 (conditional edge에서 사용)
                    # 다시 가져와야 함 (trace 데코레이터가 설정한 값)
                    current_span = _current_execution_span.get()
                    if current_span and current_span.get("id"):
                        _node_execution_spans[name] = current_span["id"]
                    
                    return result
                
                # trace 데코레이터 적용 (wrapper를 감쌈)
                traced_action = client.trace(sync_wrapper, span_kind="langgraph_node", name=name)
                traced_action._nora_traced = True
                return original_add_node(self, name, traced_action, **kwargs)
        else:
            return original_add_node(self, name, action, **kwargs)

    def traced_add_conditional_edges(
        self,
        source: str,
        path: Union[Callable, Dict[str, str]],
        path_map: Optional[Dict[str, str]] = None,
        then: Optional[str] = None,
    ):
        """Conditional edges 추가 시 routing 함수를 trace하고 decision을 기록합니다."""
        client = _get_active_client()
        
        if client and callable(path):
            # routing 함수를 trace로 감싸되 decision 정보 추가
            original_path = path
            
            @wraps(original_path)
            def traced_routing(state):
                from ...client import _current_trace_group
                
                # routing 함수 실행
                selected = original_path(state)
                
                # trace_group이 활성화된 경우에만 decision 기록
                current_group = _current_trace_group.get()
                if current_group and current_group.trace_id:
                    # path_map에서 가능한 옵션들 추출
                    if path_map:
                        options = [{"content": key, "score": None} for key in path_map.keys()]
                    else:
                        # path_map이 없으면 선택된 값으로부터 추정
                        options = [{"content": str(selected), "score": None}]
                    
                    # source 노드의 execution span ID를 부모로 사용
                    parent_execution_id = _node_execution_spans.get(source)
                    
                    # span_data 생성
                    import time
                    import uuid
                    span_data = {
                        "id": str(uuid.uuid4()),
                        "name": original_path.__name__,  # 함수 이름 사용
                        "start_time": time.time(),
                        "end_time": time.time(),
                        "span_kind": "select",
                        "status": "completed",
                        "input": state if isinstance(state, dict) else {"state": str(state)},
                        "result": {
                            "options": options,
                            "selected_option": {"content": str(selected), "score": None}
                        },
                        "metadata": {
                            "function": original_path.__name__,
                            "source_node": source,
                        },
                    }
                    
                    # 부모 execution_id 추가
                    if parent_execution_id:
                        span_data["execution_parent_id"] = parent_execution_id
                    
                    # execution span 전송
                    client._send_execution_span(current_group.trace_id, span_data)
                
                return selected
            
            # 키워드 인자로 전달
            kwargs = {}
            if path_map is not None:
                kwargs['path_map'] = path_map
            if then is not None:
                kwargs['then'] = then
            
            return original_add_conditional_edges(self, source, traced_routing, **kwargs)
        else:
            # 키워드 인자로 전달
            kwargs = {}
            if path_map is not None:
                kwargs['path_map'] = path_map
            if then is not None:
                kwargs['then'] = then
            
            return original_add_conditional_edges(self, source, path, **kwargs)

    # 메서드 패치
    StateGraph.add_node = traced_add_node
    StateGraph.add_conditional_edges = traced_add_conditional_edges
    
    # 패치 완료 마커
    StateGraph.add_node._nora_patched = True
    StateGraph.add_conditional_edges._nora_patched = True

    print("[Nora] ✓ LangGraph patched successfully (nodes + conditional edges)")



"""
Nora DATASET API
DATASET 생성 및 관리 API
"""

from typing import Optional, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import NoraClient


class Dataset_item_list:
    score: Optional[Union[float, str]]
    input: dict
    output: dict


class DataSetAPI:
    """
    DATASET 관리를 위한 API 클래스
    
    사용법:
        import nora
        
        # nora.init()으로 설정 (api_url을 DATASET API URL로도 사용)
        nora.init(
            api_key="YOUR_KEY",
            api_url="https://api.example.com/v1"
        )
        client = nora.Client()
        
        # 데이터셋 생성
        dataset = client.dataset.create(
            dataset_name="my_dataset",
            dataset_item_list=[
                {"input": {"question": "What is AI?"}, "output": {"answer": "AI is..."}, "score": 0.95},
                {"input": {"question": "What is ML?"}, "output": {"answer": "ML is..."}, "score": 0.88}
            ],
            description="QA dataset for training"
        )
        
        # 데이터셋 목록 조회
        datasets = client.dataset.list()
        
        # 특정 데이터셋 조회
        dataset_info = client.dataset.get(dataset_name="my_dataset")
    """

    def __init__(self, client: "NoraClient"):
        """
        Args:
            client: NoraClient 인스턴스
        """
        self._client = client

        # 전역 설정에서 api_url 가져오기
        from .. import _global_config
        api_url = _global_config.get("api_url")

        if api_url:
            self._api_url = api_url.rstrip("/")
        else:
            self._api_url = None

    def _get_headers(self) -> dict:
        """
        공통 헤더 생성
        """
        headers = {
            "Content-Type": "application/json",
            "X_API_KEY": f"{self._client.api_key}",
        }
        return headers
    

    def create(
        self,
        name: str,
        dataset_items: List[Dataset_item_list],
        description: Optional[str] = None,
    ) -> dict:
        """
        새로운 DATASET을 생성합니다.
        
        Args:
            name: DATASET 이름
            dataset_itme_list: DATASET 아이템 리스트
            description: DATASET 설명 (옵션)
        
        Returns:
            생성된 DATASET 정보
        """
        import requests
        if not self._api_url:
            raise ValueError(
                "DATASET API URL이 설정되지 않았습니다. "
                "nora.init(api_key='...', api_url='...')을 먼저 호출하세요."
            )
        
        url = f"{self._api_url}/dataset-sdk/"
        payload = {
            "dataset_name": name,
            "dataset_item_list": dataset_items,
            "description": description,
        }
        
        response = requests.post(
            url,
            json=payload,
            headers=self._get_headers()
        )
        response.raise_for_status()
        return response.json()
    

    def list(self) -> dict:
        """
        모든 DATASET 목록을 가져옵니다.
        
        Returns:
            DATASET 목록
        """
        import requests
        if not self._api_url:
            raise ValueError(
                "DATASET API URL이 설정되지 않았습니다. "
                "nora.init(api_key='...', api_url='...')을 먼저 호출하세요."
            )
        
        url = f"{self._api_url}/dataset-sdk/"
        
        response = requests.get(
            url,
            headers=self._get_headers()
        )
        response.raise_for_status()
        return response.json()
    
    def get(self, name: str) -> dict:
        """
        DATASET ID로 특정 DATASET 정보를 가져옵니다.
        
        Args:
            dataset_id: DATASET ID
        
        Returns:
            DATASET 정보
        """
        import requests
        if not self._api_url:
            raise ValueError(
                "DATASET API URL이 설정되지 않았습니다. "
                "nora.init(api_key='...', api_url='...')을 먼저 호출하세요."
            )
        
        url = f"{self._api_url}/dataset-sdk/{name}"
        
        response = requests.get(
            url,
            headers=self._get_headers()
        )
        response.raise_for_status()
        return response.json()


"""Gemini API 스트리밍 응답 처리"""

from typing import Any
from ...client import NoraClient
from .types import RequestParams
from .metadata_builder import build_trace_data


class StreamWrapper:
    """Generator를 context manager로 감싸는 wrapper 클래스"""
    
    def __init__(self, generator, original_response=None):
        self._generator = generator
        self._original_response = original_response
    
    def __iter__(self):
        return self._generator
    
    def __aiter__(self):
        return self._generator
    
    def __enter__(self):
        # 원본 response가 context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__enter__'):
            self._original_response.__enter__()
        return self
    
    def __exit__(self, *args):
        # 원본 response가 context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__exit__'):
            return self._original_response.__exit__(*args)
        return False
    
    async def __aenter__(self):
        # 원본 response가 async context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__aenter__'):
            await self._original_response.__aenter__()
        return self
    
    async def __aexit__(self, *args):
        # 원본 response가 async context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__aexit__'):
            return await self._original_response.__aexit__(*args)
        return False


def wrap_streaming_response(
    stream: Any,
    client: NoraClient,
    request_params: RequestParams,
    start_time: float,
    is_async: bool = False,
    parent_execution_id: str = None,
):
    """
    스트리밍 응답을 래핑하여 완료 시 trace를 기록합니다.

    Args:
        stream: 원본 스트리밍 응답
        client: Nora 클라이언트
        request_params: 요청 파라미터
        start_time: 요청 시작 시간
        is_async: 비동기 스트림 여부
        parent_execution_id: 부모 execution span ID (optional)

    Returns:
        래핑된 스트리밍 응답
    """
    # 스트리밍 데이터 수집용
    collected_data = {
        "text": "",
        "total_tokens": 0,
        "finish_reason": None,
    }

    def process_chunk(chunk):
        """청크를 처리하고 데이터를 수집합니다."""
        try:
            # 청크에서 텍스트 추출
            if hasattr(chunk, "text"):
                collected_data["text"] += chunk.text

            # 토큰 사용량 누적
            if hasattr(chunk, "usage_metadata"):
                metadata = chunk.usage_metadata
                if hasattr(metadata, "total_token_count"):
                    collected_data["total_tokens"] = metadata.total_token_count

            # finish_reason 추출
            if hasattr(chunk, "candidates") and chunk.candidates:
                candidate = chunk.candidates[0]
                if hasattr(candidate, "finish_reason"):
                    finish_reason = candidate.finish_reason
                    if hasattr(finish_reason, "name"):
                        collected_data["finish_reason"] = finish_reason.name

        except Exception:
            pass

        return chunk

    def finalize_trace():
        """스트리밍 완료 후 trace를 기록합니다."""
        import time

        end_time = time.time()

        # Mock response 객체 생성
        mock_response = type(
            "MockResponse",
            (),
            {
                "text": collected_data["text"],
                "usage_metadata": type(
                    "UsageMetadata", (), {"total_token_count": collected_data["total_tokens"]}
                )(),
                "candidates": [
                    type(
                        "Candidate",
                        (),
                        {
                            "finish_reason": type(
                                "FinishReason", (), {"name": collected_data["finish_reason"] or "STOP"}
                            )()
                        },
                    )()
                ],
            },
        )()

        trace_data = build_trace_data(
            request_params,
            mock_response,
            start_time,
            end_time,
            error=None,
        )
        
        # Add parent execution ID if exists
        if parent_execution_id:
            trace_data["execution_parent_id"] = parent_execution_id
        
        client.trace(**trace_data)

    # 동기 스트리밍 래퍼
    if not is_async:

        def sync_wrapper():
            try:
                for chunk in stream:
                    yield process_chunk(chunk)
            finally:
                finalize_trace()

        return StreamWrapper(sync_wrapper(), stream)

    # 비동기 스트리밍 래퍼
    else:

        async def async_wrapper():
            try:
                async for chunk in stream:
                    yield process_chunk(chunk)
            finally:
                finalize_trace()

        return StreamWrapper(async_wrapper(), stream)



"""
OpenAI API 관련 타입 정의
"""

from typing import TypedDict, Optional, List, Dict, Any, Literal, Union


class ToolCallInfo(TypedDict, total=False):
    """Tool call 정보"""

    id: str
    type: Literal["function"]
    function: Dict[str, str]  # name, arguments


class ResponseContent(TypedDict, total=False):
    """API 응답 내용"""

    text: str
    tool_calls: Optional[List[ToolCallInfo]]
    logprobs: Optional[Dict[str, Any]]


class UsageInfo(TypedDict, total=False):
    """토큰 사용량 정보"""

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int
    reasoning_tokens: int
    cached_tokens: int
    audio_tokens: int
    accepted_prediction_tokens: int
    rejected_prediction_tokens: int


class RequestParams(TypedDict, total=False):
    """요청 파라미터 (Chat Completions 또는 Responses API)"""

    model: str
    # Chat Completions API
    messages: List[Dict[str, Any]]
    temperature: Optional[float]
    max_tokens: Optional[int]
    max_completion_tokens: Optional[int]
    top_p: Optional[float]
    frequency_penalty: Optional[float]
    presence_penalty: Optional[float]
    n: int
    stream: bool
    stop: Optional[Union[str, List[str]]]
    tools: Optional[List[Dict[str, Any]]]
    tool_choice: Optional[Union[str, Dict[str, Any]]]
    response_format: Optional[Dict[str, Any]]
    seed: Optional[int]
    logprobs: Optional[bool]
    top_logprobs: Optional[int]
    user: Optional[str]
    # Responses API
    input: List[Dict[str, Any]]
    max_output_tokens: Optional[int]
    metadata: Optional[Dict[str, Any]]


class ToolExecutionInfo(TypedDict):
    """Tool 실행 정보"""

    call_id: str
    name: str
    arguments: Union[str, Dict[str, Any]]



"""
Anthropic 라이브러리 자동 패치
Anthropic API 호출을 자동으로 trace합니다.
동기/비동기/스트리밍 모두 지원합니다.
"""

import time
from typing import Any, Callable, Dict

from ...client import _get_active_client as get_active_client, _current_execution_span
from ..._internal.anthropic import (
    extract_request_params,
    build_trace_data,
    wrap_streaming_response,
)


def patch_anthropic() -> None:
    """Anthropic 라이브러리를 패치하여 자동 trace를 활성화합니다."""
    try:
        import anthropic
        from anthropic.resources.messages import Messages, AsyncMessages
    except ImportError:
        return  # Anthropic이 설치되지 않은 경우 무시

    # 이미 패치되었는지 확인
    if hasattr(anthropic, "_nora_patched"):
        return

    # 원본 메서드 저장
    original_sync_create = Messages.create
    original_async_create = AsyncMessages.create

    # 동기 버전 패치
    Messages.create = _create_patched_function(original_sync_create, is_async=False)

    # 비동기 버전 패치
    AsyncMessages.create = _create_patched_function(original_async_create, is_async=True)

    # 패치 완료 표시
    anthropic._nora_patched = True


def _create_patched_function(original_func: Callable, is_async: bool) -> Callable:
    """
    패치된 함수를 생성합니다.

    Args:
        original_func: 원본 함수
        is_async: 비동기 함수 여부

    Returns:
        패치된 함수
    """
    if is_async:

        async def patched_async(self, *args, **kwargs):
            return await _execute_with_trace_async(original_func, self, args, kwargs)

        return patched_async
    else:

        def patched_sync(self, *args, **kwargs):
            return _execute_with_trace_sync(original_func, self, args, kwargs)

        return patched_sync


def _execute_with_trace_sync(
    original_func: Callable, self: Any, args: tuple, kwargs: Dict[str, Any]
) -> Any:
    """
    동기 함수에서 Trace를 포함하여 원본 함수를 실행합니다.

    Args:
        original_func: 원본 함수
        self: 함수의 self 인자
        args: 위치 인자
        kwargs: 키워드 인자

    Returns:
        함수 실행 결과
    """
    client = get_active_client()
    if not client or not client.enabled:
        return original_func(self, *args, **kwargs)

    start_time = time.time()
    request_params = extract_request_params(kwargs)

    # Get parent execution span
    parent_span = _current_execution_span.get()
    parent_execution_id = parent_span.get("id") if parent_span else None

    try:
        response = original_func(self, *args, **kwargs)
        end_time = time.time()

        # 스트리밍 처리
        if kwargs.get("stream", False):
            return wrap_streaming_response(
                response, client, request_params, start_time, is_async=False, parent_execution_id=parent_execution_id
            )

        # 일반 응답 처리
        from ..._internal.anthropic import extract_response_content

        response_content = extract_response_content(response)
        trace_data = build_trace_data(request_params, response_content, start_time, end_time)
        
        # Add parent execution ID if exists
        if parent_execution_id:
            trace_data["execution_parent_id"] = parent_execution_id
        
        client._trace_method(**trace_data)
        return response

    except Exception as e:
        end_time = time.time()
        trace_data = build_trace_data(request_params, None, start_time, end_time, error=str(e))
        
        # Add parent execution ID if exists
        if parent_execution_id:
            trace_data["execution_parent_id"] = parent_execution_id
        
        client._trace_method(**trace_data)
        raise


async def _execute_with_trace_async(
    original_func: Callable, self: Any, args: tuple, kwargs: Dict[str, Any]
) -> Any:
    """
    비동기 함수에서 Trace를 포함하여 원본 함수를 실행합니다.

    Args:
        original_func: 원본 함수
        self: 함수의 self 인자
        args: 위치 인자
        kwargs: 키워드 인자

    Returns:
        함수 실행 결과
    """
    client = get_active_client()
    if not client or not client.enabled:
        return await original_func(self, *args, **kwargs)

    start_time = time.time()
    request_params = extract_request_params(kwargs)

    # Get parent execution span
    parent_span = _current_execution_span.get()
    parent_execution_id = parent_span.get("id") if parent_span else None

    try:
        response = await original_func(self, *args, **kwargs)
        end_time = time.time()

        # 스트리밍 처리
        if kwargs.get("stream", False):
            return await wrap_streaming_response(
                response, client, request_params, start_time, is_async=True, parent_execution_id=parent_execution_id
            )

        # 일반 응답 처리
        from ..._internal.anthropic import extract_response_content

        response_content = extract_response_content(response)
        trace_data = build_trace_data(request_params, response_content, start_time, end_time)
        
        # Add parent execution ID if exists
        if parent_execution_id:
            trace_data["execution_parent_id"] = parent_execution_id
        
        client._trace_method(**trace_data)
        return response

    except Exception as e:
        end_time = time.time()
        trace_data = build_trace_data(request_params, None, start_time, end_time, error=str(e))
        
        # Add parent execution ID if exists
        if parent_execution_id:
            trace_data["execution_parent_id"] = parent_execution_id
        
        client._trace_method(**trace_data)
        raise



"""
Nora Observability SDK
AI 라이브러리 호출을 자동으로 trace하는 Observability 서비스

사용법:
    import nora

    # 방법 1: init()으로 한 번만 설정
    nora.init(api_key="YOUR_KEY")
    client = nora.Client()  # api_key 자동 사용

    # 방법 2: 직접 api_key 전달
    client = nora.Client(api_key="YOUR_KEY")
    # 또는
    client = nora.NoraClient(api_key="YOUR_KEY")

    # 이제 OpenAI, Anthropic 등의 호출이 자동으로 trace됩니다!
"""

from typing import Optional, List
from .client import NoraClient

# 전역 설정 저장
_global_config: dict = {}


def init(
    api_key: str,
    api_url: str = "https://noraobservabilityback-production.up.railway.app/v1",
    thread_id: Optional[str] = None,
    user_id: Optional[str] = None,
    traced_functions: Optional[List[str]] = None,
    **kwargs,
) -> None:
    """
    Nora SDK를 전역으로 초기화합니다.
    init()으로 설정한 후에는 nora.Client()를 api_key 없이 호출할 수 있습니다.

    Args:
        api_key: Nora API 키
        api_url: Trace 데이터를 전송할 API 엔드포인트 URL (프롬프트 API도 동일하게 사용됨)
        thread_id: 스레드/세션 ID (옵션)
        user_id: 사용자 ID (옵션)
        traced_functions: 자동으로 trace_group으로 감쌀 함수 이름 리스트
        **kwargs: NoraClient의 다른 파라미터들 (batch_size, flush_interval 등)
    """
    global _global_config
    
    _global_config = {
        "api_key": api_key,
        "api_url": api_url,
        **kwargs
    }
    
    if thread_id is not None:
        _global_config["thread_id"] = thread_id
    if user_id is not None:
        _global_config["user_id"] = user_id

    # traced_functions가 있으면 별도로 저장 (클라이언트 생성 후 적용)
    if traced_functions:
        _global_config["traced_functions"] = traced_functions
    
    # 전역 클라이언트 인스턴스 생성
    global _global_client
    _global_client = NoraClient(**_global_config)
    
    # Return client instance for API compatibility
    return _global_client


# 전역 클라이언트 인스턴스
_global_client: Optional[NoraClient] = None


def trace_group(name: Optional[str] = None, metadata: Optional[dict] = None):
    """
    여러 LLM 호출을 하나의 논리적 그룹으로 묶습니다.
    
    nora.init()을 먼저 호출해야 합니다.
    
    Context manager 또는 데코레이터로 사용 가능합니다.
    
    Args:
        name: 그룹 이름 (데코레이터 사용 시 기본값: 함수 이름)
        metadata: 그룹 메타데이터
    
    예제 (Context Manager):
        >>> with nora.trace_group("multi_agent_workflow"):
        ...     response = client.chat.completions.create(...)
    
    예제 (데코레이터):
        >>> @nora.trace_group(name="batch_process")
        ... def generate():
        ...     return client.chat.completions.create(...)
    """
    if _global_client is None:
        raise RuntimeError("nora.init() must be called before using trace_group")
    return _global_client.trace_group(name=name, metadata=metadata)


def trace(func=None, span_kind: Optional[str] = None):
    """
    함수 실행을 trace합니다.
    
    nora.init()을 먼저 호출해야 합니다.
    trace_group 내에서 사용하여 개별 작업을 추적합니다.
    
    Args:
        func: 추적할 함수 (데코레이터로 사용 시)
        span_kind: span 종류
    
    예제 (Context Manager):
        >>> with nora.trace_group("workflow"):
        ...     with nora.trace(name="step1"):
        ...         result = some_function()
    
    예제 (데코레이터):
        >>> @nora.trace
        ... def my_function():
        ...     return "result"
    """
    if _global_client is None:
        raise RuntimeError("nora.init() must be called before using trace")
    return _global_client.trace(func=func, span_kind=span_kind)


# Client를 NoraClient의 별칭으로 export
Client = NoraClient

__version__ = "0.1.9"

__all__ = [
    "NoraClient",
    "Client",
    "init",
    "trace_group",
    "trace",
    "__version__",
]



"""
Anthropic API 응답/요청 처리 유틸리티
"""

from typing import Any, Dict, Optional
from ..._internal.utils import safe_extract_attr, format_messages
from .types import RequestParams, ResponseContent, UsageInfo


def extract_request_params(kwargs: Dict[str, Any]) -> RequestParams:
    """
    요청 파라미터를 추출합니다.

    Args:
        kwargs: API 호출 키워드 인자

    Returns:
        추출된 요청 파라미터
    """
    return RequestParams(
        model=kwargs.get("model", "unknown"),
        messages=kwargs.get("messages", []),
        max_tokens=kwargs.get("max_tokens"),
        temperature=kwargs.get("temperature"),
        top_p=kwargs.get("top_p"),
        top_k=kwargs.get("top_k"),
        stop_sequences=kwargs.get("stop_sequences"),
        system=kwargs.get("system"),
        stream=kwargs.get("stream", False),
        tools=kwargs.get("tools"),
        metadata=kwargs.get("metadata"),
    )


def format_prompt(request_params: RequestParams) -> str:
    """
    요청 파라미터에서 prompt를 추출합니다.

    Args:
        request_params: 요청 파라미터

    Returns:
        포맷된 prompt 문자열
    """
    # System 프롬프트 추가
    prompt_parts = []
    if request_params.system:
        prompt_parts.append(f"System: {request_params.system}")

    # 메시지 추가
    if request_params.messages:
        prompt_parts.append(format_messages(request_params.messages))

    return "\n".join(prompt_parts) if prompt_parts else ""


def extract_response_content(response: Any) -> ResponseContent:
    """
    Anthropic API 응답에서 텍스트, stop_reason 등을 추출합니다.

    Args:
        response: Anthropic API 응답 객체

    Returns:
        추출된 응답 내용
    """
    result = ResponseContent()

    try:
        # 텍스트 추출
        if hasattr(response, "content") and response.content:
            content = response.content
            if isinstance(content, list) and content:
                # Anthropic은 content가 리스트
                text_parts = []
                for block in content:
                    if hasattr(block, "text"):
                        text_parts.append(block.text)
                    elif isinstance(block, dict) and block.get("type") == "text":
                        text_parts.append(block.get("text", ""))
                result.text = "".join(text_parts)

        # Stop reason 추출
        if hasattr(response, "stop_reason"):
            result.stop_reason = response.stop_reason

        # 사용량 추출
        result.usage = extract_usage_info(response)

    except Exception:
        pass

    return result


def extract_usage_info(response: Any) -> Optional[UsageInfo]:
    """
    Anthropic 응답 객체에서 토큰 사용량을 추출합니다.

    Args:
        response: Anthropic API 응답 객체

    Returns:
        토큰 사용량 정보
    """
    try:
        usage = safe_extract_attr(response, "usage")
        if usage:
            input_tokens = safe_extract_attr(usage, "input_tokens", default=0)
            output_tokens = safe_extract_attr(usage, "output_tokens", default=0)
            return UsageInfo(input_tokens=input_tokens, output_tokens=output_tokens)
        return None
    except Exception:
        return None


def extract_tokens(response: Any) -> Optional[int]:
    """
    Anthropic 응답 객체에서 총 토큰 사용량을 반환합니다.

    Args:
        response: Anthropic API 응답 객체

    Returns:
        총 토큰 사용량 (input + output)
    """
    usage = extract_usage_info(response)
    return usage.total_tokens if usage else None



"""
Anthropic 스트리밍 응답 처리
"""

from typing import Any, Dict
from .utils import format_prompt
from .types import RequestParams


class StreamWrapper:
    """Generator를 context manager로 감싸는 wrapper 클래스"""
    
    def __init__(self, generator, original_response=None):
        self._generator = generator
        self._original_response = original_response
    
    def __iter__(self):
        return self._generator
    
    def __aiter__(self):
        return self._generator
    
    def __enter__(self):
        # 원본 response가 context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__enter__'):
            self._original_response.__enter__()
        return self
    
    def __exit__(self, *args):
        # 원본 response가 context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__exit__'):
            return self._original_response.__exit__(*args)
        return False
    
    async def __aenter__(self):
        # 원본 response가 async context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__aenter__'):
            await self._original_response.__aenter__()
        return self
    
    async def __aexit__(self, *args):
        # 원본 response가 async context manager를 지원하면 사용
        if self._original_response and hasattr(self._original_response, '__aexit__'):
            return await self._original_response.__aexit__(*args)
        return False


def wrap_streaming_response(
    response: Any,
    client: Any,
    request_params: RequestParams,
    start_time: float,
    is_async: bool,
    parent_execution_id: str = None,
) -> Any:
    """
    스트리밍 응답을 래핑하여 완료 시 trace를 기록합니다.

    Args:
        response: Anthropic 스트리밍 응답
        client: Nora client 인스턴스
        request_params: 요청 파라미터
        start_time: 시작 시간
        is_async: 비동기 여부
        parent_execution_id: 부모 execution span ID (optional)

    Returns:
        래핑된 스트리밍 응답
    """
    # 스트리밍 데이터 수집용
    collected_data: Dict[str, Any] = {
        "text": "",
        "stop_reason": None,
        "usage": None,
    }

    def process_event(event: Any) -> Any:
        """이벤트를 처리하고 데이터를 수집합니다."""
        try:
            # 이벤트 타입에 따른 처리
            event_type = getattr(event, "type", None)

            if event_type == "content_block_delta":
                # 텍스트 델타
                delta = getattr(event, "delta", None)
                if delta:
                    delta_type = getattr(delta, "type", None)
                    if delta_type == "text_delta":
                        text = getattr(delta, "text", "")
                        if text:
                            collected_data["text"] += text

            elif event_type == "message_delta":
                # Stop reason 추출
                delta = getattr(event, "delta", None)
                if delta:
                    stop_reason = getattr(delta, "stop_reason", None)
                    if stop_reason:
                        collected_data["stop_reason"] = stop_reason

                # Usage 정보 추출
                usage = getattr(event, "usage", None)
                if usage:
                    collected_data["usage"] = {
                        "input_tokens": getattr(usage, "input_tokens", 0),
                        "output_tokens": getattr(usage, "output_tokens", 0),
                    }

        except Exception:
            pass

        return event

    def finalize_trace() -> None:
        """스트리밍 완료 후 trace를 기록합니다."""
        import time
        
        end_time = time.time()

        # Usage 정보 처리
        tokens_used = None
        if collected_data["usage"]:
            usage_data = collected_data["usage"]
            tokens_used = usage_data.get("input_tokens", 0) + usage_data.get("output_tokens", 0)

        trace_data = {
            "provider": "anthropic",
            "model": request_params.model,
            "prompt": format_prompt(request_params),
            "response": collected_data["text"],
            "start_time": start_time,
            "end_time": end_time,
            "stop_reason": collected_data["stop_reason"],
            "tokens_used": tokens_used,
            "metadata": {
                "temperature": request_params.temperature,
                "max_tokens": request_params.max_tokens,
                "top_p": request_params.top_p,
                "top_k": request_params.top_k,
                "system": request_params.system,
                "messages_count": len(request_params.messages),
                "is_streaming": True,
                "usage": collected_data["usage"],
            },
        }
        
        # Add parent execution ID if exists
        if parent_execution_id:
            trace_data["execution_parent_id"] = parent_execution_id

        client.trace(**trace_data)

    # 동기 스트리밍 래퍼
    if not is_async:

        def sync_wrapper():
            try:
                for event in response:
                    yield process_event(event)
            finally:
                finalize_trace()

        return StreamWrapper(sync_wrapper(), response)

    # 비동기 스트리밍 래퍼
    else:

        async def async_wrapper():
            try:
                async for event in response:
                    yield process_event(event)
            finally:
                finalize_trace()

        return StreamWrapper(async_wrapper(), response)



"""
Execution Span 전송 모듈
Execution Span 관련 로직을 NoraClient에서 분리
"""

import threading
import queue
import time
from typing import Optional, Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import NoraClient

try:
    import requests
except ImportError:
    requests = None


def _sanitize_for_json(obj: Any) -> Any:
    """Convert Pydantic models and other non-serializable objects to JSON-serializable format."""
    if obj is None:
        return None
    
    # Handle Pydantic models
    if hasattr(obj, 'model_dump'):
        return obj.model_dump()
    elif hasattr(obj, 'dict'):
        return obj.dict()
    
    # Handle dictionaries recursively
    if isinstance(obj, dict):
        return {k: _sanitize_for_json(v) for k, v in obj.items()}
    
    # Handle lists recursively
    if isinstance(obj, (list, tuple)):
        return [_sanitize_for_json(item) for item in obj]
    
    # Handle basic types
    if isinstance(obj, (str, int, float, bool)):
        return obj
    
    # Fallback: try to convert to string
    try:
        # Check if it's JSON serializable
        import json
        json.dumps(obj)
        return obj
    except (TypeError, ValueError):
        return str(obj)


class ExecutionSpanManager:
    """Execution Span 전송을 담당하는 클래스"""

    def __init__(self, client: "NoraClient"):  # type: ignore
        """ExecutionSpanManager 초기화
        
        Args:
            client: NoraClient 인스턴스
        """
        self.client = client

        # Map trace_id -> list of execution_ids
        self._trace_execution_ids: Dict[str, list] = {}
        self._trace_exec_lock = threading.Lock()

        # Queue for ordered, asynchronous sending of execution spans
        self._span_queue: "queue.Queue[tuple[str, Dict[str, Any]]]" = queue.Queue()

        # Worker thread that sends execution spans one-by-one (preserves order)
        self._worker_thread = threading.Thread(target=self._worker, daemon=True)
        self._worker_thread.start()

        # Track active execution span threads (kept for compatibility)
        self._execution_threads: list = []
        self._execution_threads_lock = threading.Lock()

    def send_execution_span(self, trace_id: str, span_data: Dict[str, Any]) -> None:
        """Send execution span immediately to API and capture execution_id."""
        if not requests:
            print("[Nora] ⚠️  WARNING: requests module not available")
            return

        if not trace_id:
            print("[Nora] ⚠️  WARNING: trace_id is empty or None")
            return
        # Enqueue span for ordered asynchronous sending by worker thread
        # Use local id for fallback mapping in worker if needed
        try:
            self._span_queue.put_nowait((trace_id, span_data))
        except Exception:
            # If queueing fails for any reason, fall back to direct send
            self._do_send(trace_id, span_data)

    def wait_for_completion(self, timeout: float = 2.0) -> None:
        """Wait for all pending execution span threads to complete."""
        # Wait for the span queue to be processed (ordered completion)
        start = time.time()
        while True:
            if self._span_queue.empty():
                break
            if (time.time() - start) >= timeout:
                print(f"[Nora] ⚠️ Timeout waiting for execution span queue to empty")
                break
            time.sleep(0.01)
        
        # Give worker thread a bit more time to finish current item
        time.sleep(0.05)

        # Also join any auxiliary threads if present (compatibility)
        with self._execution_threads_lock:
            threads_to_wait = [t for t in self._execution_threads if t.is_alive()]

        for t in threads_to_wait:
            t.join(timeout=timeout)

        # Clean up finished threads
        with self._execution_threads_lock:
            self._execution_threads = [t for t in self._execution_threads if t.is_alive()]

    def _worker(self) -> None:
        """Worker thread that processes the span queue in FIFO order."""
        while True:
            try:
                item = self._span_queue.get()
            except Exception:
                continue

            try:
                trace_id, span_data = item
                self._do_send(trace_id, span_data)
            except Exception:
                pass
            finally:
                try:
                    self._span_queue.task_done()
                except Exception:
                    pass

    def _do_send(self, trace_id: str, span_data: Dict[str, Any]) -> None:
        """Actual sending logic extracted from previous inline _send() closure."""
        if not requests:
            print("[Nora] ⚠️  WARNING: requests module not available")
            return

        local_span_id = span_data.get("id")

        try:
            # Sanitize span_data to ensure all Pydantic models are converted to dicts
            sanitized_span_data = _sanitize_for_json(span_data)
            
            headers = {
                "X-API-Key": self.client.api_key,
                "Content-Type": "application/json",
            }

            span_name = sanitized_span_data.get("name") or (
                sanitized_span_data.get("provider", "unknown") + "_" + sanitized_span_data.get("model", "execution")
            )

            payload = {
                "trace_id": trace_id,
                "span_name": span_name,
                "span_data": sanitized_span_data,
            }
            
            # Include execution_parent_id if present
            execution_parent_id = sanitized_span_data.get("execution_parent_id")
            if execution_parent_id:
                payload["execution_parent_id"] = execution_parent_id

            # Parse and update trace input
            try:
                parsed_input = self.client._parse_input_from_span(sanitized_span_data)
                if parsed_input:
                    self.client._update_trace_input(trace_id, parsed_input)
            except Exception:
                pass

            # Include model and tokens for LLM spans
            try:
                span_kind = sanitized_span_data.get("span_kind")
            except Exception:
                span_kind = None

            if span_kind == "llm":
                model_val = sanitized_span_data.get("model")
                if model_val is not None:
                    payload["model"] = model_val

                # Build tokens object
                tokens_obj = None
                try:
                    tokens_field = sanitized_span_data.get("tokens")
                    if isinstance(tokens_field, dict):
                        if ("total_input" in tokens_field) or ("total_output" in tokens_field):
                            ti = tokens_field.get("total_input")
                            to = tokens_field.get("total_output")
                            obj: Dict[str, int] = {}
                            if ti is not None:
                                obj["total_input"] = int(ti)
                            if to is not None:
                                obj["total_output"] = int(to)
                            tokens_obj = obj if obj else None
                        else:
                            # Try alternative key names
                            alt_ti = None
                            alt_to = None
                            if isinstance(tokens_field.get("input_tokens"), int):
                                alt_ti = tokens_field.get("input_tokens")
                            if isinstance(tokens_field.get("output_tokens"), int):
                                alt_to = tokens_field.get("output_tokens")
                            if alt_ti is None and isinstance(
                                tokens_field.get("prompt_tokens"), int
                            ):
                                alt_ti = tokens_field.get("prompt_tokens")
                            if alt_to is None and isinstance(
                                tokens_field.get("completion_tokens"), int
                            ):
                                alt_to = tokens_field.get("completion_tokens")
                            if alt_to is None and isinstance(
                                tokens_field.get("tokens_used"), int
                            ):
                                alt_to = tokens_field.get("tokens_used")

                            if alt_ti is not None or alt_to is not None:
                                obj: Dict[str, int] = {}
                                if alt_ti is not None:
                                    obj["total_input"] = int(alt_ti)
                                if alt_to is not None:
                                    obj["total_output"] = int(alt_to)
                                tokens_obj = obj
                    else:
                        # Try metadata/usage fields
                        md = sanitized_span_data.get("metadata") or {}
                        usage = None
                        try:
                            usage = (md.get("response") or {}).get("usage")
                        except Exception:
                            usage = None

                        total_input = None
                        total_output = None

                        if isinstance(usage, dict):
                            if isinstance(usage.get("prompt_tokens"), int):
                                total_input = usage.get("prompt_tokens")
                            if isinstance(usage.get("completion_tokens"), int):
                                total_output = usage.get("completion_tokens")

                        # Fallback to metadata.input_tokens / metadata.output_tokens
                        if total_input is None:
                            try:
                                if isinstance(md.get("input_tokens"), int):
                                    total_input = md.get("input_tokens")
                            except Exception:
                                pass
                        if total_output is None:
                            try:
                                if isinstance(md.get("output_tokens"), int):
                                    total_output = md.get("output_tokens")
                            except Exception:
                                pass

                        if total_input is not None or total_output is not None:
                            obj: Dict[str, int] = {}
                            if total_input is not None:
                                obj["total_input"] = int(total_input)
                            if total_output is not None:
                                obj["total_output"] = int(total_output)
                            tokens_obj = obj
                        else:
                            tokens_used = sanitized_span_data.get("tokens_used")
                            if tokens_used is not None:
                                tokens_obj = {"total_output": int(tokens_used)}
                except Exception as e:
                    tokens_obj = None

                if tokens_obj is not None:
                    payload["tokens"] = tokens_obj

            response = requests.post(
                self.client.execution_span_url, json=payload, headers=headers, timeout=10
            )

            if response.status_code in (200, 201):
                try:
                    response_data = response.json()
                    execution_id = response_data.get("id")

                    if execution_id:

                        with self._trace_exec_lock:
                            if trace_id not in self._trace_execution_ids:
                                self._trace_execution_ids[trace_id] = []
                            self._trace_execution_ids[trace_id].append(execution_id)
                    else:
                        if local_span_id:
                            with self._trace_exec_lock:
                                if trace_id not in self._trace_execution_ids:
                                    self._trace_execution_ids[trace_id] = []
                                self._trace_execution_ids[trace_id].append(local_span_id)
                except Exception as e:
                    if local_span_id:
                        with self._trace_exec_lock:
                            if trace_id not in self._trace_execution_ids:
                                self._trace_execution_ids[trace_id] = []
                            self._trace_execution_ids[trace_id].append(local_span_id)

        except requests.exceptions.RequestException as e:
            print(f"[Nora] ❌ Error sending execution span: {str(e)}")
        except Exception as e:
            print(f"[Nora] ❌ Unexpected error sending execution span: {str(e)}")

